# __init__.py -- Contrib module for Dulwich
# Copyright (C) 2014 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Contributed implementations and extensions for dulwich.

This package contains various contributed modules that extend dulwich
functionality. These modules are maintained as part of dulwich but may
have additional dependencies or more specialized use cases.

Available modules:
- greenthreads: Green-threaded support for finding missing objects
- paramiko_vendor: SSH client implementation using paramiko
- release_robot: Automated release management utilities
- requests_vendor: HTTP client implementation using requests
- swift: OpenStack Swift object storage backend
"""

__all__ = []



# __init__.py -- Fast export/import functionality
# Copyright (C) 2010-2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#


"""Fast export/import functionality."""

__all__ = [
    "GitFastExporter",
    "GitImportProcessor",
    "split_email",
]

import stat
from collections.abc import Generator
from typing import TYPE_CHECKING, Any, BinaryIO

from fastimport import commands, parser, processor
from fastimport import errors as fastimport_errors

from .index import commit_tree
from .object_store import iter_tree_contents
from .objects import ZERO_SHA, Blob, Commit, ObjectID, Tag
from .refs import Ref

if TYPE_CHECKING:
    from .object_store import BaseObjectStore
    from .repo import BaseRepo


def split_email(text: bytes) -> tuple[bytes, bytes]:
    """Split email address from name.

    Args:
        text: Full name and email (e.g. b"John Doe <john@example.com>")

    Returns:
        Tuple of (name, email)
    """
    # TODO(jelmer): Dedupe this and the same functionality in
    # format_annotate_line.
    (name, email) = text.rsplit(b" <", 1)
    return (name, email.rstrip(b">"))


class GitFastExporter:
    """Generate a fast-export output stream for Git objects."""

    def __init__(self, outf: BinaryIO, store: "BaseObjectStore") -> None:
        """Initialize the fast exporter.

        Args:
            outf: Output file to write to
            store: Object store to export from
        """
        self.outf = outf
        self.store = store
        self.markers: dict[bytes, ObjectID] = {}
        self._marker_idx = 0

    def print_cmd(self, cmd: object) -> None:
        """Print a command to the output stream.

        Args:
            cmd: Command object to print
        """
        if hasattr(cmd, "__bytes__"):
            output = cmd.__bytes__()
        else:
            output = cmd.__repr__().encode("utf-8")
        self.outf.write(output + b"\n")

    def _allocate_marker(self) -> bytes:
        """Allocate a new marker.

        Returns:
            New marker as bytes
        """
        self._marker_idx += 1
        return str(self._marker_idx).encode("ascii")

    def _export_blob(self, blob: Blob) -> tuple[Any, bytes]:
        """Export a blob object.

        Args:
            blob: Blob object to export

        Returns:
            Tuple of (BlobCommand, marker)
        """
        marker = self._allocate_marker()
        self.markers[marker] = blob.id
        return (commands.BlobCommand(marker, blob.data), marker)  # type: ignore[no-untyped-call,unused-ignore]

    def emit_blob(self, blob: Blob) -> bytes:
        """Emit a blob to the output stream.

        Args:
            blob: Blob object to emit

        Returns:
            Marker for the blob
        """
        (cmd, marker) = self._export_blob(blob)
        self.print_cmd(cmd)
        return marker

    def _iter_files(
        self, base_tree: ObjectID | None, new_tree: ObjectID | None
    ) -> Generator[Any, None, None]:
        for (
            (old_path, new_path),
            (old_mode, new_mode),
            (old_hexsha, new_hexsha),
        ) in self.store.tree_changes(base_tree, new_tree):
            if new_path is None:
                if old_path is not None:
                    yield commands.FileDeleteCommand(old_path)  # type: ignore[no-untyped-call,unused-ignore]
                continue
            marker = b""
            if new_mode is not None and not stat.S_ISDIR(new_mode):
                if new_hexsha is not None:
                    blob = self.store[new_hexsha]
                    from .objects import Blob

                    if isinstance(blob, Blob):
                        marker = self.emit_blob(blob)
            if old_path != new_path and old_path is not None:
                yield commands.FileRenameCommand(old_path, new_path)  # type: ignore[no-untyped-call,unused-ignore]
            if old_mode != new_mode or old_hexsha != new_hexsha:
                prefixed_marker = b":" + marker
                assert new_mode is not None
                yield commands.FileModifyCommand(  # type: ignore[no-untyped-call,unused-ignore]
                    new_path, new_mode, prefixed_marker, None
                )

    def _export_commit(
        self, commit: Commit, ref: Ref, base_tree: ObjectID | None = None
    ) -> tuple[Any, bytes]:
        file_cmds = list(self._iter_files(base_tree, commit.tree))
        marker = self._allocate_marker()
        if commit.parents:
            from_ = commit.parents[0]
            merges = commit.parents[1:]
        else:
            from_ = None
            merges = []
        author, author_email = split_email(commit.author)
        committer, committer_email = split_email(commit.committer)
        cmd = commands.CommitCommand(  # type: ignore[no-untyped-call,unused-ignore]
            ref,
            marker,
            (author, author_email, commit.author_time, commit.author_timezone),
            (
                committer,
                committer_email,
                commit.commit_time,
                commit.commit_timezone,
            ),
            commit.message,
            from_,
            merges,
            file_cmds,
        )
        return (cmd, marker)

    def emit_commit(
        self, commit: Commit, ref: Ref, base_tree: ObjectID | None = None
    ) -> bytes:
        """Emit a commit in fast-export format.

        Args:
          commit: Commit object to export
          ref: Reference name for the commit
          base_tree: Base tree for incremental export

        Returns:
          Marker for the commit
        """
        cmd, marker = self._export_commit(commit, ref, base_tree)
        self.print_cmd(cmd)
        return marker


class GitImportProcessor(processor.ImportProcessor):  # type: ignore[misc,unused-ignore]
    """An import processor that imports into a Git repository using Dulwich."""

    # FIXME: Batch creation of objects?

    def __init__(
        self,
        repo: "BaseRepo",
        params: Any | None = None,  # noqa: ANN401
        verbose: bool = False,
        outf: BinaryIO | None = None,
    ) -> None:
        """Initialize GitImportProcessor.

        Args:
          repo: Repository to import into
          params: Import parameters
          verbose: Whether to enable verbose output
          outf: Output file for verbose messages
        """
        processor.ImportProcessor.__init__(self, params, verbose)  # type: ignore[no-untyped-call,unused-ignore]
        self.repo = repo
        self.last_commit = ZERO_SHA
        self.markers: dict[bytes, ObjectID] = {}
        self._contents: dict[bytes, tuple[int, bytes]] = {}

    def lookup_object(self, objectish: bytes) -> ObjectID:
        """Look up an object by reference or marker.

        Args:
          objectish: Object reference or marker

        Returns:
          Object ID
        """
        if objectish.startswith(b":"):
            return self.markers[objectish[1:]]
        return ObjectID(objectish)

    def import_stream(self, stream: BinaryIO) -> dict[bytes, ObjectID]:
        """Import from a fast-import stream.

        Args:
          stream: Stream to import from

        Returns:
          Dictionary of markers to object IDs
        """
        p = parser.ImportParser(stream)  # type: ignore[no-untyped-call,unused-ignore]
        self.process(p.iter_commands)  # type: ignore[no-untyped-call,unused-ignore]
        return self.markers

    def blob_handler(self, cmd: commands.BlobCommand) -> None:
        """Process a BlobCommand."""
        blob = Blob.from_string(cmd.data)
        self.repo.object_store.add_object(blob)
        if cmd.mark:
            self.markers[cmd.mark] = blob.id

    def checkpoint_handler(self, cmd: commands.CheckpointCommand) -> None:
        """Process a CheckpointCommand."""

    def commit_handler(self, cmd: commands.CommitCommand) -> None:
        """Process a CommitCommand."""
        commit = Commit()
        if cmd.author is not None:
            (author_name, author_email, author_timestamp, author_timezone) = cmd.author
        else:
            (author_name, author_email, author_timestamp, author_timezone) = (
                cmd.committer
            )
        (
            committer_name,
            committer_email,
            commit_timestamp,
            commit_timezone,
        ) = cmd.committer
        if isinstance(author_name, str):
            author_name = author_name.encode("utf-8")
        if isinstance(author_email, str):
            author_email = author_email.encode("utf-8")
        commit.author = author_name + b" <" + author_email + b">"
        commit.author_timezone = author_timezone
        commit.author_time = int(author_timestamp)
        if isinstance(committer_name, str):
            committer_name = committer_name.encode("utf-8")
        if isinstance(committer_email, str):
            committer_email = committer_email.encode("utf-8")
        commit.committer = committer_name + b" <" + committer_email + b">"
        commit.commit_timezone = commit_timezone
        commit.commit_time = int(commit_timestamp)
        commit.message = cmd.message
        commit.parents = []
        if cmd.from_:
            cmd.from_ = self.lookup_object(cmd.from_)
            self._reset_base(cmd.from_)
        for filecmd in cmd.iter_files():  # type: ignore[no-untyped-call,unused-ignore]
            if filecmd.name == b"filemodify":
                assert isinstance(filecmd, commands.FileModifyCommand)
                if filecmd.data is not None:
                    blob = Blob.from_string(filecmd.data)
                    self.repo.object_store.add_object(blob)
                    blob_id = blob.id
                else:
                    assert filecmd.dataref is not None
                    blob_id = self.lookup_object(filecmd.dataref)
                self._contents[filecmd.path] = (filecmd.mode, blob_id)
            elif filecmd.name == b"filedelete":
                assert isinstance(filecmd, commands.FileDeleteCommand)
                del self._contents[filecmd.path]
            elif filecmd.name == b"filecopy":
                assert isinstance(filecmd, commands.FileCopyCommand)
                self._contents[filecmd.dest_path] = self._contents[filecmd.src_path]
            elif filecmd.name == b"filerename":
                assert isinstance(filecmd, commands.FileRenameCommand)
                self._contents[filecmd.new_path] = self._contents[filecmd.old_path]
                del self._contents[filecmd.old_path]
            elif filecmd.name == b"filedeleteall":
                self._contents = {}
            else:
                raise Exception(f"Command {filecmd.name!r} not supported")
        from dulwich.objects import ObjectID

        commit.tree = commit_tree(
            self.repo.object_store,
            (
                (path, ObjectID(hexsha), mode)
                for (path, (mode, hexsha)) in self._contents.items()
            ),
        )
        if self.last_commit != ZERO_SHA:
            commit.parents.append(self.last_commit)
        for merge in cmd.merges:
            commit.parents.append(self.lookup_object(merge))
        self.repo.object_store.add_object(commit)
        self.repo[cmd.ref] = commit.id
        self.last_commit = commit.id
        if cmd.mark:
            mark_bytes = (
                cmd.mark
                if isinstance(cmd.mark, bytes)
                else str(cmd.mark).encode("ascii")
            )
            self.markers[mark_bytes] = commit.id

    def progress_handler(self, cmd: commands.ProgressCommand) -> None:
        """Process a ProgressCommand."""

    def _reset_base(self, commit_id: ObjectID) -> None:
        if self.last_commit == commit_id:
            return
        self._contents = {}
        self.last_commit = commit_id
        if commit_id != ZERO_SHA:
            from .objects import Commit

            commit = self.repo[commit_id]
            tree_id = commit.tree if isinstance(commit, Commit) else None
            if tree_id is None:
                return
            for (
                path,
                mode,
                hexsha,
            ) in iter_tree_contents(self.repo.object_store, tree_id):
                assert path is not None and mode is not None and hexsha is not None
                self._contents[path] = (mode, hexsha)

    def reset_handler(self, cmd: commands.ResetCommand) -> None:
        """Process a ResetCommand."""
        from_: ObjectID
        if cmd.from_ is None:
            from_ = ZERO_SHA
        else:
            from_ = self.lookup_object(cmd.from_)
        self._reset_base(from_)
        self.repo.refs[Ref(cmd.ref)] = from_

    def tag_handler(self, cmd: commands.TagCommand) -> None:
        """Process a TagCommand."""
        tag = Tag()
        tag.tagger = cmd.tagger
        tag.message = cmd.message
        tag.name = cmd.from_
        self.repo.object_store.add_object(tag)
        self.repo.refs["refs/tags/" + tag.name] = tag.id

    def feature_handler(self, cmd: commands.FeatureCommand) -> None:
        """Process a FeatureCommand."""
        feature_name = (
            cmd.feature_name.decode("utf-8")
            if isinstance(cmd.feature_name, bytes)
            else cmd.feature_name
        )
        raise fastimport_errors.UnknownFeature(feature_name)  # type: ignore[no-untyped-call,unused-ignore]



# rerere.py -- Git rerere (reuse recorded resolution) implementation
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Git rerere (reuse recorded resolution) implementation.

This module implements Git's rerere functionality, which records and reuses
manual conflict resolutions. When a merge conflict occurs, rerere records the
conflict and the resolution. If the same conflict happens again, rerere can
automatically apply the recorded resolution.
"""

__all__ = [
    "RerereCache",
    "is_rerere_autoupdate",
    "is_rerere_enabled",
    "rerere_auto",
]

import hashlib
import os
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dulwich.config import StackedConfig
    from dulwich.repo import Repo


def _compute_conflict_id(preimage: bytes) -> str:
    """Compute a unique ID for a conflict based on its preimage.

    The preimage is the conflict with all conflict markers normalized
    to a canonical form (without side names).

    Args:
        preimage: The normalized conflict content

    Returns:
        Hex digest of the SHA-1 hash
    """
    return hashlib.sha1(preimage).hexdigest()


def _normalize_conflict_markers(content: bytes) -> bytes:
    """Normalize conflict markers to a canonical form.

    This removes the side names (ours/theirs) from conflict markers
    to create a preimage that only depends on the structure of the conflict.

    Args:
        content: File content with conflict markers

    Returns:
        Normalized content with generic conflict markers
    """
    lines = content.split(b"\n")
    result = []

    for line in lines:
        # Normalize conflict start markers
        if line.startswith(b"<<<<<<<"):
            result.append(b"<<<<<<<")
        # Normalize conflict end markers
        elif line.startswith(b">>>>>>>"):
            result.append(b">>>>>>>")
        # Keep separator as is
        elif line == b"=======":
            result.append(line)
        else:
            result.append(line)

    return b"\n".join(result)


def _extract_conflict_regions(content: bytes) -> list[tuple[bytes, bytes, bytes]]:
    """Extract conflict regions from file content.

    Args:
        content: File content with conflict markers

    Returns:
        List of tuples (ours, separator, theirs) for each conflict region
    """
    conflicts = []
    lines = content.split(b"\n")
    i = 0

    while i < len(lines):
        if lines[i].startswith(b"<<<<<<<"):
            # Found conflict start
            ours_lines = []
            theirs_lines = []
            i += 1

            # Collect "ours" lines
            while i < len(lines) and not lines[i].startswith(b"======="):
                ours_lines.append(lines[i])
                i += 1

            if i >= len(lines):
                break

            # Skip separator
            i += 1

            # Collect "theirs" lines
            while i < len(lines) and not lines[i].startswith(b">>>>>>>"):
                theirs_lines.append(lines[i])
                i += 1

            if i >= len(lines):
                break

            ours = b"\n".join(ours_lines)
            theirs = b"\n".join(theirs_lines)
            conflicts.append((ours, b"=======", theirs))

        i += 1

    return conflicts


def _has_conflict_markers(content: bytes) -> bool:
    """Check if content contains conflict markers.

    Args:
        content: File content to check

    Returns:
        True if conflict markers are present
    """
    return b"<<<<<<<" in content and b"=======" in content and b">>>>>>>" in content


def _remove_conflict_markers(content: bytes) -> bytes:
    """Remove all conflict markers from resolved content.

    This is used to extract the resolution from a file that has been
    manually resolved by the user.

    Args:
        content: Resolved file content

    Returns:
        Content with conflict markers removed
    """
    lines = content.split(b"\n")
    result = []
    in_conflict = False

    for line in lines:
        if line.startswith(b"<<<<<<<"):
            in_conflict = True
            continue
        elif line.startswith(b">>>>>>>"):
            in_conflict = False
            continue
        elif line == b"=======":
            continue

        if not in_conflict:
            result.append(line)

    return b"\n".join(result)


class RerereCache:
    """Manages the rerere cache in .git/rr-cache/."""

    def __init__(self, rr_cache_dir: bytes | str) -> None:
        """Initialize RerereCache.

        Args:
            rr_cache_dir: Path to the .git/rr-cache directory
        """
        if isinstance(rr_cache_dir, bytes):
            self.rr_cache_dir = os.fsdecode(rr_cache_dir)
        else:
            self.rr_cache_dir = rr_cache_dir

    @classmethod
    def from_repo(cls, repo: "Repo") -> "RerereCache":
        """Create a RerereCache from a repository.

        Args:
            repo: A Dulwich repository object

        Returns:
            RerereCache instance for the repository
        """
        rr_cache_dir = os.path.join(repo.controldir(), "rr-cache")
        return cls(rr_cache_dir)

    def _ensure_cache_dir(self) -> None:
        """Ensure the rr-cache directory exists."""
        os.makedirs(self.rr_cache_dir, exist_ok=True)

    def _get_conflict_dir(self, conflict_id: str) -> str:
        """Get the directory path for a specific conflict.

        Args:
            conflict_id: The conflict ID (SHA-1 hash)

        Returns:
            Path to the conflict directory
        """
        return os.path.join(self.rr_cache_dir, conflict_id)

    def _get_preimage_path(self, conflict_id: str) -> str:
        """Get the path to the preimage file for a conflict.

        Args:
            conflict_id: The conflict ID

        Returns:
            Path to the preimage file
        """
        return os.path.join(self._get_conflict_dir(conflict_id), "preimage")

    def _get_postimage_path(self, conflict_id: str) -> str:
        """Get the path to the postimage file for a conflict.

        Args:
            conflict_id: The conflict ID

        Returns:
            Path to the postimage file
        """
        return os.path.join(self._get_conflict_dir(conflict_id), "postimage")

    def record_conflict(self, path: bytes, content: bytes) -> str | None:
        """Record a conflict in the rerere cache.

        Args:
            path: Path to the conflicted file
            content: File content with conflict markers

        Returns:
            The conflict ID if recorded, None if no conflict markers found
        """
        if not _has_conflict_markers(content):
            return None

        self._ensure_cache_dir()

        # Normalize conflict markers and compute ID
        preimage = _normalize_conflict_markers(content)
        conflict_id = _compute_conflict_id(preimage)

        # Create conflict directory
        conflict_dir = self._get_conflict_dir(conflict_id)
        os.makedirs(conflict_dir, exist_ok=True)

        # Write preimage
        preimage_path = self._get_preimage_path(conflict_id)
        with open(preimage_path, "wb") as f:
            f.write(content)

        return conflict_id

    def record_resolution(self, conflict_id: str, content: bytes) -> None:
        """Record a resolution for a previously recorded conflict.

        Args:
            conflict_id: The conflict ID to record resolution for
            content: Resolved file content (without conflict markers)
        """
        # Write postimage
        postimage_path = self._get_postimage_path(conflict_id)

        # Ensure directory exists
        conflict_dir = self._get_conflict_dir(conflict_id)
        os.makedirs(conflict_dir, exist_ok=True)

        with open(postimage_path, "wb") as f:
            f.write(content)

    def has_resolution(self, conflict_id: str) -> bool:
        """Check if a resolution exists for a conflict.

        Args:
            conflict_id: The conflict ID

        Returns:
            True if a postimage exists for this conflict
        """
        postimage_path = self._get_postimage_path(conflict_id)
        return os.path.exists(postimage_path)

    def get_resolution(self, conflict_id: str) -> bytes | None:
        """Get the recorded resolution for a conflict.

        Args:
            conflict_id: The conflict ID

        Returns:
            The resolution content, or None if not found
        """
        postimage_path = self._get_postimage_path(conflict_id)

        try:
            with open(postimage_path, "rb") as f:
                return f.read()
        except FileNotFoundError:
            return None

    def apply_resolution(self, conflict_id: str, content: bytes) -> bytes | None:
        """Apply a recorded resolution to current conflict content.

        Args:
            conflict_id: The conflict ID
            content: Current file content with conflict markers

        Returns:
            Resolved content, or None if resolution couldn't be applied
        """
        resolution = self.get_resolution(conflict_id)
        if resolution is None:
            return None

        # For now, return the resolution directly
        # A more sophisticated implementation would merge the resolution
        # with the current conflict
        return resolution

    def forget(self, conflict_id: str) -> None:
        """Forget a recorded conflict and its resolution.

        Args:
            conflict_id: The conflict ID to forget
        """
        conflict_dir = self._get_conflict_dir(conflict_id)

        # Remove preimage and postimage files
        for filename in ["preimage", "postimage"]:
            path = os.path.join(conflict_dir, filename)
            try:
                os.remove(path)
            except FileNotFoundError:
                pass

        # Remove conflict directory if empty
        try:
            os.rmdir(conflict_dir)
        except OSError:
            pass

    def clear(self) -> None:
        """Clear all recorded conflicts and resolutions."""
        if not os.path.exists(self.rr_cache_dir):
            return

        # Remove all conflict directories
        for entry in os.listdir(self.rr_cache_dir):
            conflict_dir = os.path.join(self.rr_cache_dir, entry)
            if os.path.isdir(conflict_dir):
                # Remove all files in the directory
                for filename in os.listdir(conflict_dir):
                    os.remove(os.path.join(conflict_dir, filename))
                os.rmdir(conflict_dir)

    def gc(self, max_age_days: int = 60) -> None:
        """Garbage collect old conflict resolutions.

        Args:
            max_age_days: Maximum age in days for keeping resolutions
        """
        if not os.path.exists(self.rr_cache_dir):
            return

        cutoff_time = time.time() - (max_age_days * 24 * 60 * 60)

        for entry in os.listdir(self.rr_cache_dir):
            conflict_dir = os.path.join(self.rr_cache_dir, entry)
            if not os.path.isdir(conflict_dir):
                continue

            postimage_path = os.path.join(conflict_dir, "postimage")

            # Only remove if postimage exists and is old
            if os.path.exists(postimage_path):
                mtime = os.path.getmtime(postimage_path)
                if mtime < cutoff_time:
                    self.forget(entry)

    def status(self) -> list[tuple[str, bool]]:
        """Get the status of all conflicts in the cache.

        Returns:
            List of tuples (conflict_id, has_resolution)
        """
        if not os.path.exists(self.rr_cache_dir):
            return []

        result = []
        for entry in os.listdir(self.rr_cache_dir):
            conflict_dir = os.path.join(self.rr_cache_dir, entry)
            if os.path.isdir(conflict_dir):
                has_res = self.has_resolution(entry)
                result.append((entry, has_res))

        return sorted(result)

    def diff(self, conflict_id: str) -> tuple[bytes | None, bytes | None]:
        """Get the preimage and postimage for a conflict.

        Args:
            conflict_id: The conflict ID

        Returns:
            Tuple of (preimage, postimage), either may be None
        """
        preimage_path = self._get_preimage_path(conflict_id)
        postimage_path = self._get_postimage_path(conflict_id)

        preimage = None
        postimage = None

        try:
            with open(preimage_path, "rb") as f:
                preimage = f.read()
        except FileNotFoundError:
            pass

        try:
            with open(postimage_path, "rb") as f:
                postimage = f.read()
        except FileNotFoundError:
            pass

        return preimage, postimage


def is_rerere_enabled(config: "StackedConfig") -> bool:
    """Check if rerere is enabled in the config.

    Args:
        config: Git configuration

    Returns:
        True if rerere is enabled
    """
    try:
        enabled = config.get((b"rerere",), b"enabled")
        if enabled is None:
            return False
        if isinstance(enabled, bytes):
            return enabled.lower() in (b"true", b"1", b"yes", b"on")
        return bool(enabled)
    except KeyError:
        return False


def is_rerere_autoupdate(config: "StackedConfig") -> bool:
    """Check if rerere.autoupdate is enabled in the config.

    Args:
        config: Git configuration

    Returns:
        True if rerere.autoupdate is enabled
    """
    try:
        autoupdate = config.get((b"rerere",), b"autoupdate")
        if autoupdate is None:
            return False
        if isinstance(autoupdate, bytes):
            return autoupdate.lower() in (b"true", b"1", b"yes", b"on")
        return bool(autoupdate)
    except KeyError:
        return False


def rerere_auto(
    repo: "Repo",
    working_tree_path: bytes | str,
    conflicts: list[bytes],
) -> tuple[list[tuple[bytes, str]], list[bytes]]:
    """Automatically record conflicts and apply known resolutions.

    This is the main entry point for rerere integration with merge operations.
    It should be called after a merge that resulted in conflicts.

    Args:
        repo: Repository object
        working_tree_path: Path to the working tree
        conflicts: List of conflicted file paths

    Returns:
        Tuple of:
        - List of tuples (path, conflict_id) for recorded conflicts
        - List of paths where resolutions were automatically applied
    """
    config = repo.get_config_stack()
    if not is_rerere_enabled(config):
        return [], []

    cache = RerereCache.from_repo(repo)
    recorded = []
    resolved = []

    if isinstance(working_tree_path, bytes):
        working_tree_path = os.fsdecode(working_tree_path)

    autoupdate = is_rerere_autoupdate(config)

    # Record conflicts from the working tree and apply known resolutions
    for path in conflicts:
        # Read the file from the working tree
        file_path = os.path.join(working_tree_path, os.fsdecode(path))

        try:
            with open(file_path, "rb") as f:
                content = f.read()
        except FileNotFoundError:
            # File was deleted in conflict
            continue

        # Record the conflict
        conflict_id = cache.record_conflict(path, content)
        if not conflict_id:
            continue

        recorded.append((path, conflict_id))

        # Check if we have a resolution for this conflict
        if autoupdate and cache.has_resolution(conflict_id):
            resolution = cache.get_resolution(conflict_id)
            if resolution is not None:
                # Apply the resolution to the working tree
                with open(file_path, "wb") as f:
                    f.write(resolution)
                resolved.append(path)

    return recorded, resolved



# trailers.py -- Git trailers parsing and manipulation
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Git trailers parsing and manipulation.

This module provides functionality for parsing and manipulating Git trailers,
which are structured information blocks appended to commit messages.

Trailers follow the format:
    Token: value
    Token: value

They are similar to RFC 822 email headers and appear at the end of commit
messages after free-form content.
"""

__all__ = [
    "Trailer",
    "add_trailer_to_message",
    "format_trailers",
    "parse_trailers",
]


class Trailer:
    """Represents a single Git trailer.

    Args:
        key: The trailer key/token (e.g., "Signed-off-by")
        value: The trailer value
        separator: The separator character used (default ':')
    """

    def __init__(self, key: str, value: str, separator: str = ":") -> None:
        """Initialize a Trailer instance.

        Args:
            key: The trailer key/token
            value: The trailer value
            separator: The separator character (default ':')
        """
        self.key = key
        self.value = value
        self.separator = separator

    def __eq__(self, other: object) -> bool:
        """Compare two Trailer instances for equality.

        Args:
            other: The object to compare with

        Returns:
            True if trailers have the same key, value, and separator
        """
        if not isinstance(other, Trailer):
            return NotImplemented
        return (
            self.key == other.key
            and self.value == other.value
            and self.separator == other.separator
        )

    def __repr__(self) -> str:
        """Return a string representation suitable for debugging.

        Returns:
            A string showing the trailer's key, value, and separator
        """
        return f"Trailer(key={self.key!r}, value={self.value!r}, separator={self.separator!r})"

    def __str__(self) -> str:
        """Return the trailer formatted as it would appear in a commit message.

        Returns:
            The trailer in the format "key: value"
        """
        return f"{self.key}{self.separator} {self.value}"


def parse_trailers(
    message: bytes,
    separators: str = ":",
) -> tuple[bytes, list[Trailer]]:
    """Parse trailers from a commit message.

    Trailers are extracted from the input by looking for a group of one or more
    lines that (i) is all trailers, or (ii) contains at least one Git-generated
    or user-configured trailer and consists of at least 25% trailers.

    The group must be preceded by one or more empty (or whitespace-only) lines.
    The group must either be at the end of the input or be the last non-whitespace
    lines before a line that starts with '---'.

    Args:
        message: The commit message as bytes
        separators: Characters to recognize as trailer separators (default ':')

    Returns:
        A tuple of (message_without_trailers, list_of_trailers)
    """
    if not message:
        return (b"", [])

    # Decode message
    try:
        text = message.decode("utf-8")
    except UnicodeDecodeError:
        text = message.decode("latin-1")

    lines = text.splitlines(keepends=True)

    # Find the trailer block by searching backwards
    # Look for a blank line followed by trailer-like lines
    trailer_start = None
    cutoff_line = None

    # First, check if there's a "---" line that marks the end of the message
    for i in range(len(lines) - 1, -1, -1):
        if lines[i].lstrip().startswith("---"):
            cutoff_line = i
            break

    # Determine the search range
    search_end = cutoff_line if cutoff_line is not None else len(lines)

    # Search backwards for the trailer block
    # A trailer block must be preceded by a blank line and extend to the end
    for i in range(search_end - 1, -1, -1):
        line = lines[i].rstrip()

        # Check if this is a blank line
        if not line:
            # Check if the lines after this blank line are trailers
            potential_trailers = lines[i + 1 : search_end]

            # Remove trailing blank lines from potential trailers
            while potential_trailers and not potential_trailers[-1].strip():
                potential_trailers = potential_trailers[:-1]

            # Check if these lines form a trailer block and extend to search_end
            if potential_trailers and _is_trailer_block(potential_trailers, separators):
                # Verify these trailers extend to the end (search_end)
                # by checking there are no non-blank lines after them
                last_trailer_index = i + 1 + len(potential_trailers)
                has_content_after = False
                for j in range(last_trailer_index, search_end):
                    if lines[j].strip():
                        has_content_after = True
                        break

                if not has_content_after:
                    trailer_start = i + 1
                    break

    if trailer_start is None:
        # No trailer block found
        return (message, [])

    # Parse the trailers
    trailer_lines = lines[trailer_start:search_end]
    trailers = _parse_trailer_lines(trailer_lines, separators)

    # Reconstruct the message without trailers
    # Keep everything before the blank line that precedes the trailers
    message_lines = lines[:trailer_start]

    # Remove trailing blank lines from the message
    while message_lines and not message_lines[-1].strip():
        message_lines.pop()

    message_without_trailers = "".join(message_lines)
    if message_without_trailers and not message_without_trailers.endswith("\n"):
        message_without_trailers += "\n"

    return (message_without_trailers.encode("utf-8"), trailers)


def _is_trailer_block(lines: list[str], separators: str) -> bool:
    """Check if a group of lines forms a valid trailer block.

    A trailer block must be composed entirely of trailer lines (with possible
    blank lines and continuation lines). A single non-trailer line invalidates
    the entire block.

    Args:
        lines: The lines to check
        separators: Valid separator characters

    Returns:
        True if the lines form a valid trailer block
    """
    if not lines:
        return False

    # Remove empty lines at the end
    while lines and not lines[-1].strip():
        lines = lines[:-1]

    if not lines:
        return False

    has_any_trailer = False

    i = 0
    while i < len(lines):
        line = lines[i].rstrip()

        if not line:
            # Empty lines are allowed within the trailer block
            i += 1
            continue

        # Check if this line is a continuation (starts with whitespace)
        if line and line[0].isspace():
            # This is a continuation of the previous line
            i += 1
            continue

        # Check if this is a trailer line
        is_trailer = False
        for sep in separators:
            if sep in line:
                key_part = line.split(sep, 1)[0]
                # Key must not contain whitespace
                if key_part and not any(c.isspace() for c in key_part):
                    is_trailer = True
                    has_any_trailer = True
                    break

        # If this is not a trailer line, the block is invalid
        if not is_trailer:
            return False

        i += 1

    # Must have at least one trailer
    return has_any_trailer


def _parse_trailer_lines(lines: list[str], separators: str) -> list[Trailer]:
    """Parse individual trailer lines.

    Args:
        lines: The trailer lines to parse
        separators: Valid separator characters

    Returns:
        List of parsed Trailer objects
    """
    trailers: list[Trailer] = []
    current_trailer: Trailer | None = None

    for line in lines:
        stripped = line.rstrip()

        if not stripped:
            # Empty line - finalize current trailer if any
            if current_trailer:
                trailers.append(current_trailer)
                current_trailer = None
            continue

        # Check if this is a continuation line (starts with whitespace)
        if stripped[0].isspace():
            if current_trailer:
                # Append to the current trailer value
                continuation = stripped.lstrip()
                current_trailer.value += " " + continuation
            continue

        # Finalize the previous trailer if any
        if current_trailer:
            trailers.append(current_trailer)
            current_trailer = None

        # Try to parse as a new trailer
        for sep in separators:
            if sep in stripped:
                parts = stripped.split(sep, 1)
                key = parts[0]

                # Key must not contain whitespace
                if key and not any(c.isspace() for c in key):
                    value = parts[1].strip() if len(parts) > 1 else ""
                    current_trailer = Trailer(key, value, sep)
                    break

    # Don't forget the last trailer
    if current_trailer:
        trailers.append(current_trailer)

    return trailers


def format_trailers(trailers: list[Trailer]) -> bytes:
    """Format a list of trailers as bytes.

    Args:
        trailers: List of Trailer objects

    Returns:
        Formatted trailers as bytes
    """
    if not trailers:
        return b""

    lines = [str(trailer) for trailer in trailers]
    return "\n".join(lines).encode("utf-8") + b"\n"


def add_trailer_to_message(
    message: bytes,
    key: str,
    value: str,
    separator: str = ":",
    where: str = "end",
    if_exists: str = "addIfDifferentNeighbor",
    if_missing: str = "add",
) -> bytes:
    """Add a trailer to a commit message.

    Args:
        message: The original commit message
        key: The trailer key
        value: The trailer value
        separator: The separator to use
        where: Where to add the trailer ('end', 'start', 'after', 'before')
        if_exists: How to handle existing trailers with the same key
            - 'add': Always add
            - 'replace': Replace all existing
            - 'addIfDifferent': Add only if value is different from all existing
            - 'addIfDifferentNeighbor': Add only if value differs from neighbors
            - 'doNothing': Don't add if key exists
        if_missing: What to do if the key doesn't exist
            - 'add': Add the trailer
            - 'doNothing': Don't add the trailer

    Returns:
        The message with the trailer added
    """
    message_body, existing_trailers = parse_trailers(message, separator)

    new_trailer = Trailer(key, value, separator)

    # Check if the key exists
    key_exists = any(t.key == key for t in existing_trailers)

    if not key_exists:
        if if_missing == "doNothing":
            return message
        # Add the new trailer
        updated_trailers = [*existing_trailers, new_trailer]
    else:
        # Key exists - apply if_exists logic
        if if_exists == "doNothing":
            return message
        elif if_exists == "replace":
            # Replace all trailers with this key
            updated_trailers = [t for t in existing_trailers if t.key != key]
            updated_trailers.append(new_trailer)
        elif if_exists == "addIfDifferent":
            # Add only if no existing trailer has the same value
            has_same_value = any(
                t.key == key and t.value == value for t in existing_trailers
            )
            if has_same_value:
                return message
            updated_trailers = [*existing_trailers, new_trailer]
        elif if_exists == "addIfDifferentNeighbor":
            # Add only if adjacent trailers with same key have different values
            should_add = True

            # Check if there's a neighboring trailer with the same key and value
            for i, t in enumerate(existing_trailers):
                if t.key == key and t.value == value:
                    # Check if it's a neighbor (last trailer with this key)
                    is_neighbor = True
                    for j in range(i + 1, len(existing_trailers)):
                        if existing_trailers[j].key == key:
                            is_neighbor = False
                            break
                    if is_neighbor:
                        should_add = False
                        break

            if not should_add:
                return message
            updated_trailers = [*existing_trailers, new_trailer]
        else:  # 'add'
            updated_trailers = [*existing_trailers, new_trailer]

    # Apply where logic
    if where == "start":
        updated_trailers = [new_trailer] + [
            t for t in updated_trailers if t != new_trailer
        ]
    elif where == "before":
        # Insert before the first trailer with the same key
        result = []
        inserted = False
        for t in updated_trailers:
            if not inserted and t.key == key and t != new_trailer:
                result.append(new_trailer)
                inserted = True
            if t != new_trailer:
                result.append(t)
        if not inserted:
            result.append(new_trailer)
        updated_trailers = result
    elif where == "after":
        # Insert after the last trailer with the same key
        result = []
        last_key_index = -1
        for i, t in enumerate(updated_trailers):
            if t.key == key and t != new_trailer:
                last_key_index = len(result)
            if t != new_trailer:
                result.append(t)

        if last_key_index >= 0:
            result.insert(last_key_index + 1, new_trailer)
        else:
            result.append(new_trailer)
        updated_trailers = result
    # 'end' is the default - trailer is already at the end

    # Reconstruct the message
    result_message = message_body
    if result_message and not result_message.endswith(b"\n"):
        result_message += b"\n"

    if updated_trailers:
        result_message += b"\n"
        result_message += format_trailers(updated_trailers)

    return result_message



# swift.py -- Repo implementation atop OpenStack SWIFT
# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Author: Fabien Boucher <fabien.boucher@enovance.com>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Repo implementation atop OpenStack SWIFT."""

__all__ = [
    "PackInfoMissingObjectFinder",
    "SwiftConnector",
    "SwiftException",
    "SwiftInfoRefsContainer",
    "SwiftObjectStore",
    "SwiftPack",
    "SwiftPackData",
    "SwiftPackReader",
    "SwiftRepo",
    "SwiftSystemBackend",
    "cmd_daemon",
    "cmd_init",
    "load_conf",
    "load_pack_info",
    "main",
    "pack_info_create",
    "swift_load_pack_index",
]

# TODO: Refactor to share more code with dulwich/repo.py.
# TODO(fbo): Second attempt to _send() must be notified via real log
# TODO(fbo): More logs for operations

import json
import logging
import os
import posixpath
import stat
import sys
import tempfile
import urllib.parse as urlparse
import zlib
from collections.abc import Callable, Iterator, Mapping
from configparser import ConfigParser
from io import BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, cast

if TYPE_CHECKING:
    from dulwich.object_format import ObjectFormat

from geventhttpclient import HTTPClient

from ..file import _GitFile
from ..lru_cache import LRUSizeCache
from ..object_store import INFODIR, PACKDIR, PackBasedObjectStore
from ..objects import S_ISGITLINK, Blob, Commit, ObjectID, Tag, Tree
from ..pack import (
    ObjectContainer,
    Pack,
    PackData,
    PackIndex,
    PackIndexer,
    PackStreamCopier,
    _compute_object_size,
    compute_file_sha,
    iter_sha1,
    load_pack_index_file,
    read_pack_header,
    unpack_object,
    write_pack_header,
    write_pack_index_v2,
    write_pack_object,
)
from ..protocol import TCP_GIT_PORT, split_peeled_refs, write_info_refs
from ..refs import HEADREF, Ref, RefsContainer, read_info_refs
from ..repo import OBJECTDIR, BaseRepo
from ..server import Backend, BackendRepo, TCPGitServer
from .greenthreads import GreenThreadsMissingObjectFinder

"""
# Configuration file sample
[swift]
# Authentication URL (Keystone or Swift)
auth_url = http://127.0.0.1:5000/v2.0
# Authentication version to use
auth_ver = 2
# The tenant and username separated by a semicolon
username = admin;admin
# The user password
password = pass
# The Object storage region to use (auth v2) (Default RegionOne)
region_name = RegionOne
# The Object storage endpoint URL to use (auth v2) (Default internalURL)
endpoint_type = internalURL
# Concurrency to use for parallel tasks (Default 10)
concurrency = 10
# Size of the HTTP pool (Default 10)
http_pool_length = 10
# Timeout delay for HTTP connections (Default 20)
http_timeout = 20
# Chunk size to read from pack (Bytes) (Default 12228)
chunk_length = 12228
# Cache size (MBytes) (Default 20)
cache_length = 20
"""


class PackInfoMissingObjectFinder(GreenThreadsMissingObjectFinder):
    """Find missing objects required for pack generation."""

    def next(self) -> tuple[bytes, int, bytes | None] | None:
        """Get the next missing object.

        Returns:
          Tuple of (sha, pack_type_num, name) or None if no more objects
        """
        while True:
            if not self.objects_to_send:
                return None
            (sha, name, leaf, _) = self.objects_to_send.pop()
            if sha not in self.sha_done:
                break
        if not leaf:
            try:
                obj = self.object_store[sha]
                if isinstance(obj, Commit):
                    self.add_todo([(obj.tree, b"", None, False)])
                elif isinstance(obj, Tree):
                    tree_items = [
                        (
                            item.sha,
                            item.path
                            if isinstance(item.path, bytes)
                            else item.path.encode("utf-8")
                            if item.path is not None
                            else b"",
                            None,
                            False,
                        )
                        for item in obj.items()
                        if item.sha is not None
                    ]
                    self.add_todo(tree_items)
                elif isinstance(obj, Tag):
                    self.add_todo([(obj.object[1], None, None, False)])
                if sha in self._tagged:
                    self.add_todo([(self._tagged[sha], None, None, True)])
            except KeyError:
                pass
        self.sha_done.add(sha)
        self.progress(f"counting objects: {len(self.sha_done)}\r".encode())
        return (
            sha,
            0,
            name if isinstance(name, bytes) else name.encode("utf-8") if name else None,
        )


def load_conf(path: str | None = None, file: str | None = None) -> ConfigParser:
    """Load configuration in global var CONF.

    Args:
      path: The path to the configuration file
      file: If provided read instead the file like object
    """
    conf = ConfigParser()
    if file:
        conf.read_file(file, path)
    else:
        confpath = None
        if not path:
            try:
                confpath = os.environ["DULWICH_SWIFT_CFG"]
            except KeyError as exc:
                raise Exception("You need to specify a configuration file") from exc
        else:
            confpath = path
        if not os.path.isfile(confpath):
            raise Exception(f"Unable to read configuration file {confpath}")
        conf.read(confpath)
    return conf


def swift_load_pack_index(
    scon: "SwiftConnector", filename: str, object_format: ObjectFormat
) -> "PackIndex":
    """Read a pack index file from Swift.

    Args:
      scon: a `SwiftConnector` instance
      filename: Path to the index file objectise
      object_format: Object format for this pack
    Returns: a `PackIndexer` instance
    """
    f = scon.get_object(filename)
    if f is None:
        raise Exception(f"Could not retrieve index file {filename}")
    if isinstance(f, bytes):
        f = BytesIO(f)
    return load_pack_index_file(filename, f, object_format)


def pack_info_create(pack_data: "PackData", pack_index: "PackIndex") -> bytes:
    """Create pack info file contents.

    Args:
      pack_data: The pack data object
      pack_index: The pack index object

    Returns:
      Compressed JSON bytes containing pack information
    """
    pack = Pack.from_objects(pack_data, pack_index)
    info: dict[bytes, Any] = {}
    for obj in pack.iterobjects():
        # Commit
        if obj.type_num == Commit.type_num:
            commit_obj = obj
            assert isinstance(commit_obj, Commit)
            info[obj.id] = (obj.type_num, commit_obj.parents, commit_obj.tree)
        # Tree
        elif obj.type_num == Tree.type_num:
            tree_obj = obj
            assert isinstance(tree_obj, Tree)
            shas = [
                (s, n, not stat.S_ISDIR(m))
                for n, m, s in tree_obj.items()
                if m is not None and not S_ISGITLINK(m)
            ]
            info[obj.id] = (obj.type_num, shas)
        # Blob
        elif obj.type_num == Blob.type_num:
            info[obj.id] = (obj.type_num,)
        # Tag
        elif obj.type_num == Tag.type_num:
            tag_obj = obj
            assert isinstance(tag_obj, Tag)
            info[obj.id] = (obj.type_num, tag_obj.object[1])
    return zlib.compress(json.dumps(info).encode("utf-8"))


def load_pack_info(
    filename: str,
    scon: "SwiftConnector | None" = None,
    file: BinaryIO | None = None,
) -> dict[str, Any] | None:
    """Load pack info from Swift or file.

    Args:
      filename: The pack info filename
      scon: Optional Swift connector to use for loading
      file: Optional file object to read from instead

    Returns:
      Dictionary containing pack information or None if not found
    """
    if not file:
        if scon is None:
            return None
        obj = scon.get_object(filename)
        if obj is None:
            return None
        if isinstance(obj, bytes):
            return cast(dict[str, Any], json.loads(zlib.decompress(obj)))
        else:
            f: BinaryIO = obj
    else:
        f = file
    try:
        return cast(dict[str, Any], json.loads(zlib.decompress(f.read())))
    finally:
        if hasattr(f, "close"):
            f.close()


class SwiftException(Exception):
    """Exception raised for Swift-related errors."""


class SwiftConnector:
    """A Connector to swift that manage authentication and errors catching."""

    def __init__(self, root: str, conf: ConfigParser) -> None:
        """Initialize a SwiftConnector.

        Args:
          root: The swift container that will act as Git bare repository
          conf: A ConfigParser Object
        """
        self.conf = conf
        self.auth_ver = self.conf.get("swift", "auth_ver")
        if self.auth_ver not in ["1", "2"]:
            raise NotImplementedError("Wrong authentication version use either 1 or 2")
        self.auth_url = self.conf.get("swift", "auth_url")
        self.user = self.conf.get("swift", "username")
        self.password = self.conf.get("swift", "password")
        self.concurrency = self.conf.getint("swift", "concurrency") or 10
        self.http_timeout = self.conf.getint("swift", "http_timeout") or 20
        self.http_pool_length = self.conf.getint("swift", "http_pool_length") or 10
        self.region_name = self.conf.get("swift", "region_name") or "RegionOne"
        self.endpoint_type = self.conf.get("swift", "endpoint_type") or "internalURL"
        self.cache_length = self.conf.getint("swift", "cache_length") or 20
        self.chunk_length = self.conf.getint("swift", "chunk_length") or 12228
        self.root = root
        block_size = 1024 * 12  # 12KB
        if self.auth_ver == "1":
            self.storage_url, self.token = self.swift_auth_v1()
        else:
            self.storage_url, self.token = self.swift_auth_v2()

        token_header = {"X-Auth-Token": str(self.token)}
        self.httpclient = HTTPClient.from_url(
            str(self.storage_url),
            concurrency=self.http_pool_length,
            block_size=block_size,
            connection_timeout=self.http_timeout,
            network_timeout=self.http_timeout,
            headers=token_header,
        )
        self.base_path = str(
            posixpath.join(urlparse.urlparse(self.storage_url).path, self.root)
        )

    def swift_auth_v1(self) -> tuple[str, str]:
        """Authenticate with Swift using v1 authentication.

        Returns:
          Tuple of (storage_url, auth_token)

        Raises:
          SwiftException: If authentication fails
        """
        self.user = self.user.replace(";", ":")
        auth_httpclient = HTTPClient.from_url(
            self.auth_url,
            connection_timeout=self.http_timeout,
            network_timeout=self.http_timeout,
        )
        headers = {"X-Auth-User": self.user, "X-Auth-Key": self.password}
        path = urlparse.urlparse(self.auth_url).path

        ret = auth_httpclient.request("GET", path, headers=headers)

        # Should do something with redirections (301 in my case)

        if ret.status_code < 200 or ret.status_code >= 300:
            raise SwiftException(
                "AUTH v1.0 request failed on "
                + f"{self.auth_url} with error code {ret.status_code} ({ret.items()!s})"
            )
        storage_url = ret["X-Storage-Url"]
        token = ret["X-Auth-Token"]
        return storage_url, token

    def swift_auth_v2(self) -> tuple[str, str]:
        """Authenticate with Swift using v2 authentication.

        Returns:
          Tuple of (storage_url, auth_token)

        Raises:
          SwiftException: If authentication fails
        """
        self.tenant, self.user = self.user.split(";")
        auth_dict = {}
        auth_dict["auth"] = {
            "passwordCredentials": {
                "username": self.user,
                "password": self.password,
            },
            "tenantName": self.tenant,
        }
        auth_json = json.dumps(auth_dict)
        headers = {"Content-Type": "application/json"}
        auth_httpclient = HTTPClient.from_url(
            self.auth_url,
            connection_timeout=self.http_timeout,
            network_timeout=self.http_timeout,
        )
        path = urlparse.urlparse(self.auth_url).path
        if not path.endswith("tokens"):
            path = posixpath.join(path, "tokens")
        ret = auth_httpclient.request("POST", path, body=auth_json, headers=headers)

        if ret.status_code < 200 or ret.status_code >= 300:
            raise SwiftException(
                "AUTH v2.0 request failed on "
                + f"{str(auth_httpclient.get_base_url()) + path} with error code {ret.status_code} ({ret.items()!s})"
            )
        auth_ret_json = json.loads(ret.read())
        token = auth_ret_json["access"]["token"]["id"]
        catalogs = auth_ret_json["access"]["serviceCatalog"]
        object_store = next(
            o_store for o_store in catalogs if o_store["type"] == "object-store"
        )
        endpoints = object_store["endpoints"]
        endpoint = next(
            endp for endp in endpoints if endp["region"] == self.region_name
        )
        return endpoint[self.endpoint_type], token

    def test_root_exists(self) -> bool | None:
        """Check that Swift container exist.

        Returns: True if exist or None it not
        """
        ret = self.httpclient.request("HEAD", self.base_path)
        if ret.status_code == 404:
            return None
        if ret.status_code < 200 or ret.status_code > 300:
            raise SwiftException(
                f"HEAD request failed with error code {ret.status_code}"
            )
        return True

    def create_root(self) -> None:
        """Create the Swift container.

        Raises:
          SwiftException: if unable to create
        """
        if not self.test_root_exists():
            ret = self.httpclient.request("PUT", self.base_path)
            if ret.status_code < 200 or ret.status_code > 300:
                raise SwiftException(
                    f"PUT request failed with error code {ret.status_code}"
                )

    def get_container_objects(self) -> list[dict[str, Any]] | None:
        """Retrieve objects list in a container.

        Returns: A list of dict that describe objects
                 or None if container does not exist
        """
        qs = "?format=json"
        path = self.base_path + qs
        ret = self.httpclient.request("GET", path)
        if ret.status_code == 404:
            return None
        if ret.status_code < 200 or ret.status_code > 300:
            raise SwiftException(
                f"GET request failed with error code {ret.status_code}"
            )
        content = ret.read()
        return cast(list[dict[str, Any]], json.loads(content))

    def get_object_stat(self, name: str) -> dict[str, Any] | None:
        """Retrieve object stat.

        Args:
          name: The object name
        Returns:
          A dict that describe the object or None if object does not exist
        """
        path = self.base_path + "/" + name
        ret = self.httpclient.request("HEAD", path)
        if ret.status_code == 404:
            return None
        if ret.status_code < 200 or ret.status_code > 300:
            raise SwiftException(
                f"HEAD request failed with error code {ret.status_code}"
            )
        resp_headers = {}
        for header, value in ret.items():
            resp_headers[header.lower()] = value
        return resp_headers

    def put_object(self, name: str, content: BinaryIO) -> None:
        """Put an object.

        Args:
          name: The object name
          content: A file object
        Raises:
          SwiftException: if unable to create
        """
        content.seek(0)
        data = content.read()
        path = self.base_path + "/" + name
        headers = {"Content-Length": str(len(data))}

        def _send() -> object:
            ret = self.httpclient.request("PUT", path, body=data, headers=headers)
            return ret

        try:
            # Sometime got Broken Pipe - Dirty workaround
            ret = _send()
        except (BrokenPipeError, ConnectionError):
            # Second attempt work
            ret = _send()

        if ret.status_code < 200 or ret.status_code > 300:  # type: ignore
            raise SwiftException(
                f"PUT request failed with error code {ret.status_code}"  # type: ignore
            )

    def get_object(self, name: str, range: str | None = None) -> bytes | BytesIO | None:
        """Retrieve an object.

        Args:
          name: The object name
          range: A string range like "0-10" to
                 retrieve specified bytes in object content
        Returns:
          A file like instance or bytestring if range is specified
        """
        headers = {}
        if range:
            headers["Range"] = f"bytes={range}"
        path = self.base_path + "/" + name
        ret = self.httpclient.request("GET", path, headers=headers)
        if ret.status_code == 404:
            return None
        if ret.status_code < 200 or ret.status_code > 300:
            raise SwiftException(
                f"GET request failed with error code {ret.status_code}"
            )
        content = cast(bytes, ret.read())

        if range:
            return content
        return BytesIO(content)

    def del_object(self, name: str) -> None:
        """Delete an object.

        Args:
          name: The object name
        Raises:
          SwiftException: if unable to delete
        """
        path = self.base_path + "/" + name
        ret = self.httpclient.request("DELETE", path)
        if ret.status_code < 200 or ret.status_code > 300:
            raise SwiftException(
                f"DELETE request failed with error code {ret.status_code}"
            )

    def del_root(self) -> None:
        """Delete the root container by removing container content.

        Raises:
          SwiftException: if unable to delete
        """
        objects = self.get_container_objects()
        if objects:
            for obj in objects:
                self.del_object(obj["name"])
        ret = self.httpclient.request("DELETE", self.base_path)
        if ret.status_code < 200 or ret.status_code > 300:
            raise SwiftException(
                f"DELETE request failed with error code {ret.status_code}"
            )


class SwiftPackReader:
    """A SwiftPackReader that mimic read and sync method.

    The reader allows to read a specified amount of bytes from
    a given offset of a Swift object. A read offset is kept internally.
    The reader will read from Swift a specified amount of data to complete
    its internal buffer. chunk_length specify the amount of data
    to read from Swift.
    """

    def __init__(self, scon: SwiftConnector, filename: str, pack_length: int) -> None:
        """Initialize a SwiftPackReader.

        Args:
          scon: a `SwiftConnector` instance
          filename: the pack filename
          pack_length: The size of the pack object
        """
        self.scon = scon
        self.filename = filename
        self.pack_length = pack_length
        self.offset = 0
        self.base_offset = 0
        self.buff = b""
        self.buff_length = self.scon.chunk_length

    def _read(self, more: bool = False) -> None:
        if more:
            self.buff_length = self.buff_length * 2
        offset = self.base_offset
        r = min(self.base_offset + self.buff_length, self.pack_length)
        ret = self.scon.get_object(self.filename, range=f"{offset}-{r}")
        if ret is None:
            self.buff = b""
        elif isinstance(ret, bytes):
            self.buff = ret
        else:
            self.buff = ret.read()

    def read(self, length: int) -> bytes:
        """Read a specified amount of Bytes form the pack object.

        Args:
          length: amount of bytes to read
        Returns:
          a bytestring
        """
        end = self.offset + length
        if self.base_offset + end > self.pack_length:
            data = self.buff[self.offset :]
            self.offset = end
            return data
        if end > len(self.buff):
            # Need to read more from swift
            self._read(more=True)
            return self.read(length)
        data = self.buff[self.offset : end]
        self.offset = end
        return data

    def seek(self, offset: int) -> None:
        """Seek to a specified offset.

        Args:
          offset: the offset to seek to
        """
        self.base_offset = offset
        self._read()
        self.offset = 0

    def read_checksum(self) -> bytes:
        """Read the checksum from the pack.

        Returns: the checksum bytestring
        """
        ret = self.scon.get_object(self.filename, range="-20")
        if ret is None:
            return b""
        elif isinstance(ret, bytes):
            return ret
        else:
            return ret.read()


class SwiftPackData(PackData):
    """The data contained in a packfile.

    We use the SwiftPackReader to read bytes from packs stored in Swift
    using the Range header feature of Swift.
    """

    def __init__(
        self,
        scon: SwiftConnector,
        filename: str | os.PathLike[str],
        object_format: "ObjectFormat | None" = None,
    ) -> None:
        """Initialize a SwiftPackReader.

        Args:
          scon: a `SwiftConnector` instance
          filename: the pack filename
          object_format: Object format for this pack
        """
        from dulwich.object_format import DEFAULT_OBJECT_FORMAT

        if object_format is None:
            import warnings

            warnings.warn(
                "SwiftPackData() should be called with object_format parameter",
                DeprecationWarning,
                stacklevel=2,
            )
            object_format = DEFAULT_OBJECT_FORMAT
        self.object_format = object_format
        self.scon = scon
        self._filename = filename
        self._header_size = 12
        headers = self.scon.get_object_stat(str(self._filename))
        if headers is None:
            raise Exception(f"Could not get stats for {self._filename}")
        self.pack_length = int(headers["content-length"])
        pack_reader = SwiftPackReader(self.scon, str(self._filename), self.pack_length)
        (_version, self._num_objects) = read_pack_header(pack_reader.read)
        self._offset_cache = LRUSizeCache(
            1024 * 1024 * self.scon.cache_length,
            compute_size=_compute_object_size,
        )
        self.pack = None

    def get_object_at(
        self, offset: int
    ) -> tuple[int, tuple[bytes | int, list[bytes]] | list[bytes]]:
        """Get the object at a specific offset in the pack.

        Args:
          offset: The offset in the pack file

        Returns:
          Tuple of (pack_type_num, object_data)
        """
        if offset in self._offset_cache:
            return self._offset_cache[offset]
        assert offset >= self._header_size
        pack_reader = SwiftPackReader(self.scon, str(self._filename), self.pack_length)
        pack_reader.seek(offset)
        unpacked, _ = unpack_object(pack_reader.read, self.object_format.hash_func)
        obj_data = unpacked._obj()
        return (unpacked.pack_type_num, obj_data)

    def get_stored_checksum(self) -> bytes:
        """Get the stored checksum for this pack.

        Returns:
          The pack checksum as bytes
        """
        pack_reader = SwiftPackReader(self.scon, str(self._filename), self.pack_length)
        return pack_reader.read_checksum()

    def close(self) -> None:
        """Close the pack data (no-op for Swift)."""


class SwiftPack(Pack):
    """A Git pack object.

    Same implementation as pack.Pack except that _idx_load and
    _data_load are bounded to Swift version of load_pack_index and
    PackData.
    """

    def __init__(self, *args: object, **kwargs: object) -> None:
        """Initialize SwiftPack.

        Args:
          *args: Arguments to pass to parent class
          **kwargs: Keyword arguments, must include 'scon' (SwiftConnector)
        """
        self.scon: SwiftConnector = kwargs["scon"]  # type: ignore
        del kwargs["scon"]
        super().__init__(*args, **kwargs)  # type: ignore
        self._pack_info_path = self._basename + ".info"
        self._pack_info: dict[str, Any] | None = None
        self._pack_info_load: Callable[[], dict[str, Any] | None] = (
            lambda: load_pack_info(self._pack_info_path, self.scon)
        )
        self._idx_load = lambda: swift_load_pack_index(
            self.scon, self._idx_path, self.object_format
        )
        self._data_load = lambda: SwiftPackData(self.scon, self._data_path)

    @property
    def pack_info(self) -> dict[str, Any] | None:
        """The pack data object being used."""
        if self._pack_info is None:
            self._pack_info = self._pack_info_load()
        return self._pack_info


class SwiftObjectStore(PackBasedObjectStore):
    """A Swift Object Store.

    Allow to manage a bare Git repository from Openstack Swift.
    This object store only supports pack files and not loose objects.
    """

    def __init__(self, scon: SwiftConnector) -> None:
        """Open a Swift object store.

        Args:
          scon: A `SwiftConnector` instance
        """
        super().__init__()
        self.scon = scon
        self.root = self.scon.root
        self.pack_dir = posixpath.join(OBJECTDIR, PACKDIR)
        self._alternates = None

    def _update_pack_cache(self) -> list[Any]:
        objects = self.scon.get_container_objects()
        if objects is None:
            return []
        pack_files = [
            o["name"].replace(".pack", "")
            for o in objects
            if o["name"].endswith(".pack")
        ]
        ret = []
        for basename in pack_files:
            pack = SwiftPack(basename, object_format=self.object_format, scon=self.scon)
            self._pack_cache[basename] = pack
            ret.append(pack)
        return ret

    def _iter_loose_objects(self) -> Iterator[Any]:
        """Loose objects are not supported by this repository."""
        return iter([])

    def pack_info_get(self, sha: ObjectID) -> tuple[Any, ...] | None:
        """Get pack info for a specific SHA.

        Args:
          sha: The SHA to look up

        Returns:
          Pack info tuple or None if not found
        """
        for pack in self.packs:
            if sha in pack:
                if hasattr(pack, "pack_info"):
                    pack_info = pack.pack_info
                    if pack_info is not None:
                        return cast(tuple[Any, ...] | None, pack_info.get(sha))
        return None

    def _collect_ancestors(
        self, heads: list[Any], common: set[Any] | None = None
    ) -> tuple[set[Any], set[Any]]:
        if common is None:
            common = set()

        def _find_parents(commit: ObjectID) -> list[Any]:
            for pack in self.packs:
                if commit in pack:
                    try:
                        if hasattr(pack, "pack_info"):
                            pack_info = pack.pack_info
                            if pack_info is not None:
                                return cast(list[Any], pack_info[commit][1])
                    except KeyError:
                        # Seems to have no parents
                        return []
            return []

        bases = set()
        commits = set()
        queue = []
        queue.extend(heads)
        while queue:
            e = queue.pop(0)
            if e in common:
                bases.add(e)
            elif e not in commits:
                commits.add(e)
                parents = _find_parents(e)
                queue.extend(parents)
        return (commits, bases)

    def add_pack(self) -> tuple[BytesIO, Callable[[], None], Callable[[], None]]:
        """Add a new pack to this object store.

        Returns: Fileobject to write to and a commit function to
            call when the pack is finished.
        """
        f = BytesIO()

        def commit() -> "SwiftPack | None":
            """Commit the pack to Swift storage.

            Returns:
              The created SwiftPack or None if empty
            """
            f.seek(0)
            from typing import cast

            from ..file import _GitFile

            pack = PackData(
                file=cast(_GitFile, f), filename="", object_format=self.object_format
            )
            entries = pack.sorted_entries()
            if entries:
                basename = posixpath.join(
                    self.pack_dir,
                    f"pack-{iter_sha1(entry[0] for entry in entries).decode('ascii')}",
                )
                index = BytesIO()
                write_pack_index_v2(index, entries, pack.get_stored_checksum())
                self.scon.put_object(basename + ".pack", f)
                f.close()
                self.scon.put_object(basename + ".idx", index)
                index.close()
                final_pack = SwiftPack(
                    basename, object_format=self.object_format, scon=self.scon
                )
                final_pack.check_length_and_checksum()
                self._add_cached_pack(basename, final_pack)
                return final_pack
            else:
                return None

        def abort() -> None:
            """Abort the pack operation (no-op)."""

        def commit_wrapper() -> None:
            """Wrapper that discards the return value."""
            commit()

        return f, commit_wrapper, abort

    def add_object(self, obj: object) -> None:
        """Add a single object to the store.

        Args:
          obj: The object to add
        """
        self.add_objects(
            [
                (obj, None),  # type: ignore
            ]
        )

    def _pack_cache_stale(self) -> bool:
        return False

    def _get_loose_object(self, sha: bytes) -> None:
        return None

    def add_thin_pack(
        self, read_all: Callable[[int], bytes], read_some: Callable[[int], bytes]
    ) -> "SwiftPack":
        """Read a thin pack.

        Read it from a stream and complete it in a temporary file.
        Then the pack and the corresponding index file are uploaded to Swift.
        """
        fd, path = tempfile.mkstemp(prefix="tmp_pack_")
        f = os.fdopen(fd, "w+b")
        try:
            pack_data = PackData(
                file=cast(_GitFile, f), filename=path, object_format=self.object_format
            )
            indexer = PackIndexer(
                cast(BinaryIO, pack_data._file),
                self.object_format.hash_func,
                resolve_ext_ref=None,
            )
            copier = PackStreamCopier(
                self.object_format.hash_func, read_all, read_some, f, delta_iter=None
            )
            copier.verify()
            return self._complete_thin_pack(f, path, copier, indexer)
        finally:
            f.close()
            os.unlink(path)

    def _complete_thin_pack(
        self, f: BinaryIO, path: str, copier: object, indexer: object
    ) -> "SwiftPack":
        entries = list(indexer)  # type: ignore

        # Update the header with the new number of objects.
        f.seek(0)
        write_pack_header(f, len(entries) + len(indexer.ext_refs()))  # type: ignore

        # Must flush before reading (http://bugs.python.org/issue3207)
        f.flush()

        # Rescan the rest of the pack, computing the SHA with the new header.
        new_sha = compute_file_sha(
            f,
            hash_func=self.object_format.hash_func,
            end_ofs=-self.object_format.oid_length,
        )

        # Must reposition before writing (http://bugs.python.org/issue3207)
        f.seek(0, os.SEEK_CUR)

        # Complete the pack.
        for ext_sha in indexer.ext_refs():  # type: ignore
            assert len(ext_sha) in (20, 32)  # SHA-1 or SHA-256
            type_num, data = self.get_raw(ext_sha)
            offset = f.tell()
            crc32 = write_pack_object(f, type_num, data, sha=new_sha)  # type: ignore
            entries.append((ext_sha, offset, crc32))
        pack_sha = new_sha.digest()
        f.write(pack_sha)
        f.flush()

        # Move the pack in.
        entries.sort()
        pack_base_name = posixpath.join(
            self.pack_dir,
            "pack-" + os.fsdecode(iter_sha1(e[0] for e in entries)),
        )
        self.scon.put_object(pack_base_name + ".pack", f)

        # Write the index.
        filename = pack_base_name + ".idx"
        index_file = BytesIO()
        write_pack_index_v2(index_file, entries, pack_sha)
        self.scon.put_object(filename, index_file)

        # Write pack info.
        f.seek(0)
        pack_data = PackData(
            filename="", file=cast(_GitFile, f), object_format=self.object_format
        )
        index_file.seek(0)
        pack_index = load_pack_index_file("", index_file, self.object_format)
        serialized_pack_info = pack_info_create(pack_data, pack_index)
        f.close()
        index_file.close()
        pack_info_file = BytesIO(serialized_pack_info)
        filename = pack_base_name + ".info"
        self.scon.put_object(filename, pack_info_file)
        pack_info_file.close()

        # Add the pack to the store and return it.
        final_pack = SwiftPack(
            pack_base_name, object_format=self.object_format, scon=self.scon
        )
        final_pack.check_length_and_checksum()
        self._add_cached_pack(pack_base_name, final_pack)
        return final_pack


class SwiftInfoRefsContainer(RefsContainer):
    """Manage references in info/refs object."""

    def __init__(self, scon: SwiftConnector, store: object) -> None:
        """Initialize SwiftInfoRefsContainer.

        Args:
          scon: Swift connector instance
          store: Object store instance
        """
        self.scon = scon
        self.filename = "info/refs"
        self.store = store
        f = self.scon.get_object(self.filename)
        if not f:
            f = BytesIO(b"")
        elif isinstance(f, bytes):
            f = BytesIO(f)

        # Initialize refs from info/refs file
        self._refs: dict[Ref, ObjectID] = {}
        self._peeled: dict[Ref, ObjectID] = {}
        refs = read_info_refs(f)
        (self._refs, self._peeled) = split_peeled_refs(refs)

    def _load_check_ref(
        self, name: Ref, old_ref: ObjectID | None
    ) -> dict[Ref, ObjectID] | bool:
        self._check_refname(name)
        obj = self.scon.get_object(self.filename)
        if not obj:
            return {}
        if isinstance(obj, bytes):
            f = BytesIO(obj)
        else:
            f = obj
        refs = read_info_refs(f)
        (refs, _peeled) = split_peeled_refs(refs)
        if old_ref is not None:
            if refs[name] != old_ref:
                return False
        return refs

    def _write_refs(self, refs: Mapping[Ref, ObjectID]) -> None:
        f = BytesIO()
        f.writelines(write_info_refs(refs, cast("ObjectContainer", self.store)))
        self.scon.put_object(self.filename, f)

    def set_if_equals(
        self,
        name: Ref,
        old_ref: ObjectID | None,
        new_ref: ObjectID,
        committer: bytes | None = None,
        timestamp: float | None = None,
        timezone: int | None = None,
        message: bytes | None = None,
    ) -> bool:
        """Set a refname to new_ref only if it currently equals old_ref."""
        if name == HEADREF:
            return True
        refs = self._load_check_ref(name, old_ref)
        if not isinstance(refs, dict):
            return False
        refs[name] = new_ref
        self._write_refs(refs)
        self._refs[name] = new_ref
        return True

    def remove_if_equals(
        self,
        name: Ref,
        old_ref: ObjectID | None,
        committer: object = None,
        timestamp: object = None,
        timezone: object = None,
        message: object = None,
    ) -> bool:
        """Remove a refname only if it currently equals old_ref."""
        if name == HEADREF:
            return True
        refs = self._load_check_ref(name, old_ref)
        if not isinstance(refs, dict):
            return False
        del refs[name]
        self._write_refs(refs)
        del self._refs[name]
        return True

    def read_loose_ref(self, name: Ref) -> bytes | None:
        """Read a loose reference."""
        return self._refs.get(name, None)

    def get_packed_refs(self) -> dict[Ref, ObjectID]:
        """Get packed references."""
        return {}

    def get_peeled(self, name: Ref) -> ObjectID | None:
        """Get peeled version of a reference."""
        try:
            return self._peeled[name]
        except KeyError:
            ref_value = self._refs.get(name)
            # Only return if it's an ObjectID (not a symref)
            if isinstance(ref_value, bytes) and len(ref_value) == 40:
                return ObjectID(ref_value)
            return None

    def allkeys(self) -> set[Ref]:
        """Get all reference names.

        Returns:
          Set of reference names as Ref
        """
        try:
            self._refs[HEADREF] = self._refs[Ref(b"refs/heads/master")]
        except KeyError:
            pass
        return set(self._refs.keys())


class SwiftRepo(BaseRepo):
    """A Git repository backed by Swift object storage."""

    def __init__(self, root: str, conf: ConfigParser) -> None:
        """Init a Git bare Repository on top of a Swift container.

        References are managed in info/refs objects by
        `SwiftInfoRefsContainer`. The root attribute is the Swift
        container that contain the Git bare repository.

        Args:
          root: The container which contains the bare repo
          conf: A ConfigParser object
        """
        self.root = root.lstrip("/")
        self.conf = conf
        self.scon = SwiftConnector(self.root, self.conf)
        objects = self.scon.get_container_objects()
        if not objects:
            raise Exception(f"There is not any GIT repo here : {self.root}")
        object_names = [o["name"].split("/")[0] for o in objects]
        if OBJECTDIR not in object_names:
            raise Exception(f"This repository ({self.root}) is not bare.")
        self.bare = True
        self._controldir = self.root
        object_store = SwiftObjectStore(self.scon)
        refs = SwiftInfoRefsContainer(self.scon, object_store)
        BaseRepo.__init__(self, object_store, refs)

    def _determine_file_mode(self) -> bool:
        """Probe the file-system to determine whether permissions can be trusted.

        Returns: True if permissions can be trusted, False otherwise.
        """
        return False

    def _put_named_file(self, filename: str, contents: bytes) -> None:
        """Put an object in a Swift container.

        Args:
          filename: the path to the object to put on Swift
          contents: the content as bytestring
        """
        with BytesIO() as f:
            f.write(contents)
            self.scon.put_object(filename, f)

    @classmethod
    def init_bare(cls, scon: SwiftConnector, conf: ConfigParser) -> "SwiftRepo":
        """Create a new bare repository.

        Args:
          scon: a `SwiftConnector` instance
          conf: a ConfigParser object
        Returns:
          a `SwiftRepo` instance
        """
        scon.create_root()
        for obj in [
            posixpath.join(OBJECTDIR, PACKDIR),
            posixpath.join(INFODIR, "refs"),
        ]:
            scon.put_object(obj, BytesIO(b""))
        ret = cls(scon.root, conf)
        ret._init_files(True)
        return ret


class SwiftSystemBackend(Backend):
    """Backend for serving Git repositories from Swift."""

    def __init__(self, logger: "logging.Logger", conf: ConfigParser) -> None:
        """Initialize SwiftSystemBackend.

        Args:
          logger: Logger instance
          conf: Configuration parser instance
        """
        self.conf = conf
        self.logger = logger

    def open_repository(self, path: str) -> "BackendRepo":
        """Open a repository at the given path.

        Args:
          path: Path to the repository in Swift

        Returns:
          SwiftRepo instance
        """
        self.logger.info("opening repository at %s", path)
        return cast("BackendRepo", SwiftRepo(path, self.conf))


def cmd_daemon(args: list[str]) -> None:
    """Start a TCP git server for Swift repositories.

    Args:
      args: Command line arguments
    """
    import optparse

    parser = optparse.OptionParser()
    parser.add_option(
        "-l",
        "--listen_address",
        dest="listen_address",
        default="127.0.0.1",
        help="Binding IP address.",
    )
    parser.add_option(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=TCP_GIT_PORT,
        help="Binding TCP port.",
    )
    parser.add_option(
        "-c",
        "--swift_config",
        dest="swift_config",
        default="",
        help="Path to the configuration file for Swift backend.",
    )
    options, args = parser.parse_args(args)

    try:
        import gevent
        import geventhttpclient  # noqa: F401
    except ImportError:
        print(
            "gevent and geventhttpclient libraries are mandatory "
            " for use the Swift backend."
        )
        sys.exit(1)
    import gevent.monkey

    gevent.monkey.patch_socket()
    from dulwich import log_utils

    logger = log_utils.getLogger(__name__)
    conf = load_conf(options.swift_config)
    backend = SwiftSystemBackend(logger, conf)

    log_utils.default_logging_config()
    server = TCPGitServer(backend, options.listen_address, port=options.port)
    server.serve_forever()


def cmd_init(args: list[str]) -> None:
    """Initialize a new Git repository in Swift.

    Args:
      args: Command line arguments
    """
    import optparse

    parser = optparse.OptionParser()
    parser.add_option(
        "-c",
        "--swift_config",
        dest="swift_config",
        default="",
        help="Path to the configuration file for Swift backend.",
    )
    options, args = parser.parse_args(args)

    conf = load_conf(options.swift_config)
    if args == []:
        parser.error("missing repository name")
    repo = args[0]
    scon = SwiftConnector(repo, conf)
    SwiftRepo.init_bare(scon, conf)


def main(argv: list[str] = sys.argv) -> None:
    """Main entry point for Swift Git command line interface.

    Args:
      argv: Command line arguments
    """
    commands = {
        "init": cmd_init,
        "daemon": cmd_daemon,
    }

    if len(argv) < 2:
        print("Usage: {} <{}> [OPTIONS...]".format(argv[0], "|".join(commands.keys())))
        sys.exit(1)

    cmd = argv[1]
    if cmd not in commands:
        print(f"No such subcommand: {cmd}")
        sys.exit(1)
    commands[cmd](argv[2:])


if __name__ == "__main__":
    main()



# config.py - Reading and writing Git config files
# Copyright (C) 2011-2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Working with Git submodules."""

__all__ = [
    "ensure_submodule_placeholder",
    "iter_cached_submodules",
]

import os
from collections.abc import Iterator
from typing import TYPE_CHECKING

from .object_store import iter_tree_contents
from .objects import S_ISGITLINK

if TYPE_CHECKING:
    from .objects import ObjectID
    from .pack import ObjectContainer
    from .repo import Repo


def iter_cached_submodules(
    store: "ObjectContainer", root_tree_id: "ObjectID"
) -> Iterator[tuple[bytes, "ObjectID"]]:
    """Iterate over cached submodules.

    Args:
      store: Object store to iterate
      root_tree_id: SHA of root tree

    Returns:
      Iterator over over (path, sha) tuples
    """
    for entry in iter_tree_contents(store, root_tree_id):
        assert entry.mode is not None
        if S_ISGITLINK(entry.mode):
            assert entry.path is not None
            assert entry.sha is not None
            yield entry.path, entry.sha


def ensure_submodule_placeholder(
    repo: "Repo",
    submodule_path: str | bytes,
) -> None:
    """Create a submodule placeholder directory with .git file.

    This creates the minimal structure needed for a submodule:
    - The submodule directory
    - A .git file pointing to the submodule's git directory

    Args:
      repo: Parent repository
      submodule_path: Path to the submodule relative to repo root
    """
    # Ensure path is bytes
    if isinstance(submodule_path, str):
        submodule_path = submodule_path.encode()

    # Get repo path as bytes
    repo_path = repo.path if isinstance(repo.path, bytes) else repo.path.encode()

    # Create full path to submodule
    full_path = os.path.join(repo_path, submodule_path)

    # Create submodule directory if it doesn't exist
    if not os.path.exists(full_path):
        os.makedirs(full_path)

    # Create .git file pointing to the submodule's git directory
    git_filename = b".git" if isinstance(full_path, bytes) else ".git"
    git_file_path = os.path.join(full_path, git_filename)
    if not os.path.exists(git_file_path):
        # Submodule git directories are typically stored in .git/modules/<name>
        # The relative path from the submodule to the parent's .git directory
        # depends on the submodule's depth
        depth = submodule_path.count(b"/") + 1
        relative_git_dir = b"../" * depth + b".git/modules/" + submodule_path

        with open(git_file_path, "wb") as f:
            f.write(b"gitdir: " + relative_git_dir + b"\n")



# aiohttp.py -- aiohttp smart client/server
# Copyright (C) 2022 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""aiohttp client/server support."""

import asyncio
import sys
from io import BytesIO
from typing import BinaryIO, cast

from aiohttp import web

from .. import log_utils
from ..errors import HangupException
from ..objects import ObjectID
from ..protocol import ReceivableProtocol
from ..repo import Repo
from ..server import (
    DEFAULT_HANDLERS,
    BackendRepo,
    DictBackend,
    generate_info_refs,
    generate_objects_info_packs,
)
from ..web import NO_CACHE_HEADERS, cache_forever_headers

logger = log_utils.getLogger(__name__)

# Application keys for type-safe access to app state
REPO_KEY = web.AppKey("repo", Repo)
HANDLERS_KEY = web.AppKey("handlers", dict)
DUMB_KEY = web.AppKey("dumb", bool)


async def send_file(
    req: web.Request, f: BinaryIO | None, headers: dict[str, str]
) -> web.StreamResponse:
    """Send a file-like object to the request output.

    Args:
      req: The HTTPGitRequest object to send output to.
      f: An open file-like object to send; will be closed.
      headers: Headers to send
    Returns: Iterator over the contents of the file, as chunks.
    """
    if f is None:
        raise web.HTTPNotFound(text="File not found")
    response = web.StreamResponse(status=200, reason="OK", headers=headers)
    await response.prepare(req)
    try:
        while True:
            data = f.read(10240)
            if not data:
                break
            await response.write(data)
    except OSError:
        raise web.HTTPInternalServerError(text="Error reading file")
    finally:
        f.close()
    await response.write_eof()
    return response


async def get_loose_object(request: web.Request) -> web.Response:
    """Handle request for a loose object.

    Args:
      request: aiohttp request object
    Returns: Response with the loose object data
    """
    sha = ObjectID(
        (request.match_info["dir"] + request.match_info["file"]).encode("ascii")
    )
    logger.info("Sending loose object %s", sha)
    object_store = request.app[REPO_KEY].object_store
    if not object_store.contains_loose(sha):
        raise web.HTTPNotFound(text="Object not found")
    try:
        data = object_store[sha].as_legacy_object()
    except OSError:
        raise web.HTTPInternalServerError(text="Error reading object")
    headers = {"Content-Type": "application/x-git-loose-object"}
    headers.update(cache_forever_headers())
    return web.Response(status=200, headers=headers, body=data)


async def get_text_file(request: web.Request) -> web.StreamResponse:
    """Handle request for a text file.

    Args:
      request: aiohttp request object
    Returns: Response with the text file contents
    """
    headers = {"Content-Type": "text/plain"}
    headers.update(NO_CACHE_HEADERS)
    path = request.match_info["file"]
    logger.info("Sending plain text file %s", path)
    repo = request.app[REPO_KEY]
    return await send_file(request, repo.get_named_file(path), headers)


async def refs_request(
    repo: Repo, request: web.Request, handlers: dict[bytes, type] | None = None
) -> web.StreamResponse | web.Response:
    """Handle a refs request.

    Args:
      repo: Repository object
      request: aiohttp request object
      handlers: Optional dict of service handlers
    Returns: Response with refs information
    """
    service = request.query.get("service")
    if service:
        if handlers is None:
            handlers = dict(DEFAULT_HANDLERS)
        handler_cls = handlers.get(service.encode("ascii"), None)
        if handler_cls is None:
            raise web.HTTPForbidden(text="Unsupported service")
        headers = {"Content-Type": f"application/x-{service}-advertisement"}
        headers.update(NO_CACHE_HEADERS)

        response = web.StreamResponse(status=200, headers=headers)

        await response.prepare(request)

        out = BytesIO()
        proto = ReceivableProtocol(BytesIO().read, out.write)
        handler = handler_cls(
            DictBackend({b".": cast(BackendRepo, repo)}),
            [b"."],
            proto,
            stateless_rpc=True,
            advertise_refs=True,
        )
        handler.proto.write_pkt_line(b"# service=" + service.encode("ascii") + b"\n")
        handler.proto.write_pkt_line(None)

        # TODO(jelmer): Implement this with proper async code
        await asyncio.to_thread(handler.handle)

        await response.write(out.getvalue())

        await response.write_eof()

        return response
    else:
        # non-smart fallback
        headers = {"Content-Type": "text/plain"}
        headers.update(NO_CACHE_HEADERS)
        logger.info("Emulating dumb info/refs")
        return web.Response(body=b"".join(generate_info_refs(repo)), headers=headers)


async def get_info_refs(request: web.Request) -> web.StreamResponse | web.Response:
    """Handle request for /info/refs.

    Args:
      request: aiohttp request object
    Returns: Response with refs information
    """
    repo = request.app[REPO_KEY]
    return await refs_request(repo, request, request.app[HANDLERS_KEY])


async def get_info_packs(request: web.Request) -> web.Response:
    """Handle request for /info/packs.

    Args:
      request: aiohttp request object
    Returns: Response with pack information
    """
    headers = {"Content-Type": "text/plain"}
    headers.update(NO_CACHE_HEADERS)
    logger.info("Emulating dumb info/packs")
    return web.Response(
        body=b"".join(generate_objects_info_packs(request.app[REPO_KEY])),
        headers=headers,
    )


async def get_pack_file(request: web.Request) -> web.StreamResponse:
    """Handle request for a pack file.

    Args:
      request: aiohttp request object
    Returns: Response with the pack file data
    """
    headers = {"Content-Type": "application/x-git-packed-objects"}
    headers.update(cache_forever_headers())
    sha = request.match_info["sha"]
    path = f"objects/pack/pack-{sha}.pack"
    logger.info("Sending pack file %s", path)
    return await send_file(
        request,
        request.app[REPO_KEY].get_named_file(path),
        headers=headers,
    )


async def get_index_file(request: web.Request) -> web.StreamResponse:
    """Handle request for a pack index file.

    Args:
      request: aiohttp request object
    Returns: Response with the pack index file data
    """
    headers = {"Content-Type": "application/x-git-packed-objects-toc"}
    headers.update(cache_forever_headers())
    sha = request.match_info["sha"]
    path = f"objects/pack/pack-{sha}.idx"
    logger.info("Sending pack file %s", path)
    return await send_file(
        request, request.app["repo"].get_named_file(path), headers=headers
    )


async def service_request(
    repo: Repo, request: web.Request, handlers: dict[bytes, type] | None = None
) -> web.StreamResponse:
    """Handle a git service request (upload-pack or receive-pack).

    Args:
      repo: Repository object
      request: aiohttp request object
      handlers: Optional dict of service handlers
    Returns: Response with service result
    """
    service = request.match_info["service"]
    if handlers is None:
        handlers = dict(DEFAULT_HANDLERS)
    logger.info("Handling service request for %s", service)
    handler_cls = handlers.get(service.encode("ascii"), None)
    if handler_cls is None:
        raise web.HTTPForbidden(text="Unsupported service")
    headers = {"Content-Type": f"application/x-{service}-result"}
    headers.update(NO_CACHE_HEADERS)

    response = web.StreamResponse(status=200, headers=headers)

    await response.prepare(request)

    inf = BytesIO(await request.read())
    outf = BytesIO()

    def handle() -> None:
        proto = ReceivableProtocol(inf.read, outf.write)
        handler = handler_cls(
            DictBackend({b".": cast(BackendRepo, repo)}),
            [b"."],
            proto,
            stateless_rpc=True,
        )
        try:
            handler.handle()
        except HangupException:
            response.force_close()

    # TODO(jelmer): Implement this with proper async code
    await asyncio.to_thread(handle)

    await response.write(outf.getvalue())

    await response.write_eof()
    return response


async def handle_service_request(request: web.Request) -> web.StreamResponse:
    """Handle a service request endpoint.

    Args:
      request: aiohttp request object
    Returns: Response with service result
    """
    repo = request.app[REPO_KEY]

    return await service_request(repo, request, request.app[HANDLERS_KEY])


def create_repo_app(
    repo: Repo, handlers: dict[bytes, type] | None = None, dumb: bool = False
) -> web.Application:
    """Create an aiohttp application for serving a git repository.

    Args:
      repo: Repository object to serve
      handlers: Optional dict of service handlers
      dumb: Whether to enable dumb HTTP protocol support
    Returns: Configured aiohttp Application
    """
    app = web.Application()
    app[REPO_KEY] = repo
    if handlers is None:
        handlers = dict(DEFAULT_HANDLERS)
    app[HANDLERS_KEY] = handlers
    app[DUMB_KEY] = dumb
    app.router.add_get("/info/refs", get_info_refs)
    app.router.add_post(
        "/{service:git-upload-pack|git-receive-pack}", handle_service_request
    )
    if dumb:
        app.router.add_get("/{file:HEAD}", get_text_file)
        app.router.add_get("/{file:objects/info/alternates}", get_text_file)
        app.router.add_get("/{file:objects/info/http-alternates}", get_text_file)
        app.router.add_get("/objects/info/packs", get_info_packs)
        app.router.add_get(
            "/objects/{dir:[0-9a-f]{2}}/{file:[0-9a-f]{38}}", get_loose_object
        )
        app.router.add_get(
            "/objects/pack/pack-{sha:[0-9a-f]{40}}\\.pack", get_pack_file
        )
        app.router.add_get(
            "/objects/pack/pack-{sha:[0-9a-f]{40}}\\.idx", get_index_file
        )
    return app


def main(argv: list[str] | None = None) -> None:
    """Entry point for starting an HTTP git server."""
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--listen_address",
        dest="listen_address",
        default="localhost",
        help="Binding IP address.",
    )
    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=8000,
        help="Port to listen on.",
    )
    parser.add_argument("gitdir", type=str, default=".", nargs="?")
    args = parser.parse_args(argv)

    log_utils.default_logging_config()
    app = create_repo_app(Repo(args.gitdir))
    logger.info(
        "Listening for HTTP connections on %s:%d",
        args.listen_address,
        args.port,
    )
    web.run_app(app, port=args.port, host=args.listen_address)


if __name__ == "__main__":
    main(sys.argv[1:])
